#ifndef PICOLED_BUFFER_H
#define PICOLED_BUFFER_H

#include <cstddef>

namespace picoled
{

template <typename T>
class buffer
{
public:
  // behavior when out of bounds
  enum class ReadMode
  {
    Default, // use default constructor for T
    Wrap,    // use modular arithmetic to put back in bounds
    Clamp    // pick from the corresponding edge
  };

  buffer(size_t width, size_t height) :
    width(width),
    height(height),
    data(new T[width * height])
  {}
  ~buffer()
  {
    delete[] data;
  }

  buffer(const buffer&) = delete;
  buffer(buffer&&) = delete;

  buffer& operator=(const buffer&) = delete;
  buffer& operator=(buffer&&) = delete;

  size_t get_width() const { return width; }
  size_t get_height() const { return height; }

  void write(int x, int y, const T& pixel)
  {
    if (x < 0 || x >= width || y < 0 || y >= height) { return; }

    data[x + y * width] = pixel;
  }

  T read(int x, int y, ReadMode mode = ReadMode::Default)
  {
    switch (mode)
    {
    case ReadMode::Default:
      return read_default(x, y);

    case ReadMode::Wrap:
      return read_wrap(x, y);

    case ReadMode::Clamp:
      return read_clamp(x, y);
    }

    return T();
  }

  void clear()
  {
    for (int i = 0; i < width * height; i++)
    {
      data[i] = T();
    }
  }

private:
  size_t width;
  size_t height;
  T* data;

  T read_default(int x, int y)
  {
    if (x < 0 || x >= width || y < 0 || y >= height) { return T(); }
    return data[x + y * width];
  }

  T read_wrap(int x, int y)
  {
    if (x < 0)
    {
      x = width - ((-x) % width);
    }
    else if (x >= width)
    {
      x = x % width;
    }

    if (y < 0)
    {
      y = height - ((-y) % height);
    }
    else if (y >= height)
    {
      y = y % height;
    }

    return data[x + y * width];
  }

  T read_clamp(int x, int y)
  {
    if (x < 0) { x = 0; }
    else if (x >= width) { x = width - 1; }

    if (y < 0) { y = 0; }
    else if (y >= height) { y = height - 1; }

    return data[x + y * width];
  }
};

}

#endif//PICOLED_BUFFER_H
